"""
define class PdPeak
"""
__author__ = 'ikibalin'
__version__ = "2019_09_06"
import os
import numpy


from pycifstar import Global


class PdPeak(object):
    """
    This section contains peak information extracted from the
    measured or, if present, the processed diffractogram. Each
    peak in this table will have a unique label (see _pd_peak_id).
    The reflections and phases associated with each peak will be
    specified in other sections (see the _pd_refln_ and
    _pd_phase_ sections).

    Note that peak positions are customarily determined from the
    processed diffractogram and thus corrections for position
    and intensity will have been previously applied.

    Example:

    loop_
    _pd_peak_index_h
    _pd_peak_index_k
    _pd_peak_index_l
    _pd_peak_index_mult
    _pd_peak_2theta
    _pd_peak_intensity_up
    _pd_peak_intensity_down
    _pd_peak_width_2theta

    reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_pd.dic/Cpd_peak.html
    """
    def __init__(self, h=[], k=[], l=[], mult=[], ttheta=[], up=[], down=[], width_2theta=[]):
        super(PdPeak, self).__init__()
        self.__pd_peak_index_h = None
        self.__pd_peak_index_k = None
        self.__pd_peak_index_l = None
        self.__pd_peak_index_mult = None
        self.__pd_peak_2theta = None
        self.__pd_peak_intensity_up = None
        self.__pd_peak_intensity_down = None
        self.__pd_peak_width_2theta = None

        self.h = h
        self.k = k
        self.l = l
        self.mult = mult
        self.ttheta = ttheta
        self.up = up
        self.down = down
        self.width_2theta = width_2theta

    @property
    def h(self):
        return self.__pd_peak_index_h
    @h.setter
    def h(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = int(round(x))
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__pd_peak_index_h = np_x_in

    @property
    def k(self):
        return self.__pd_peak_index_k
    @k.setter
    def k(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = int(round(x))
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__pd_peak_index_k = np_x_in

    @property
    def l(self):
        return self.__pd_peak_index_l
    @l.setter
    def l(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = int(round(x))
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__pd_peak_index_l = np_x_in

    @property
    def mult(self):
        return self.__pd_peak_index_mult
    @mult.setter
    def mult(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = int(round(x))
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__pd_peak_index_mult = np_x_in

    @property
    def ttheta(self):
        return self.__pd_peak_2theta
    @ttheta.setter
    def ttheta(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_peak_2theta = np_x_in


    @property
    def up(self):
        return self.__pd_peak_intensity_up
    @up.setter
    def up(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_peak_intensity_up = np_x_in

    @property
    def down(self):
        return self.__pd_peak_intensity_down
    @down.setter
    def down(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_peak_intensity_down = np_x_in

    @property
    def width_2theta(self):
        return self.__pd_peak_width_2theta
    @width_2theta.setter
    def width_2theta(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_peak_width_2theta = np_x_in

    def __repr__(self):
        ls_out = ["PdPeak:"]
        ls_out.append("    h     k     l  mult    ttheta           up         down width_2theta")
        for _1, _2, _3, _4, _5, _6, _7, _8 in zip(self.h, self.k, self.l, self.mult, self.ttheta, self.up, self.down, self.width_2theta):
            ls_out.append("{:5} {:5} {:5} {:5} {:9.2f} {:12.5f} {:12.5f} {:12.5f}".format(_1, _2, _3, _4, _5, _6, _7, _8))
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_pd_peak_index_h")
            ls_out.append("_pd_peak_index_k")
            ls_out.append("_pd_peak_index_l")
            ls_out.append("_pd_peak_mult")
            ls_out.append("_pd_peak_2theta")
            ls_out.append("_pd_peak_intensity_up")
            ls_out.append("_pd_peak_intensity_down")
            ls_out.append("_pd_peak_width_2theta")
            for _1, _2, _3, _4, _5, _6, _7, _8 in zip(self.h, self.k, self.l, self.mult, self.ttheta, self.up, self.down, self.width_2theta):
                ls_out.append("{:} {:} {:} {:} {:.3f} {:.5f} {:.5f} {:.3f}".format(_1, _2, _3, _4, _5, _6, _7, _8))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_pd_peak")
        if flag:
            cif_loop = cif_global["_pd_peak"]
            l_name = cif_loop.names
            if "_pd_peak_index_h" in l_name:
                self.h = [int(_1) for _1 in cif_loop["_pd_peak_index_h"]]
            if "_pd_peak_index_k" in l_name:
                self.k = [int(_1) for _1 in cif_loop["_pd_peak_index_k"]]
            if "_pd_peak_index_l" in l_name:
                self.l = [int(_1) for _1 in cif_loop["_pd_peak_index_l"]]
            if "_pd_peak_index_mult" in l_name:
                self.mult = [int(_1) for _1 in cif_loop["_pd_peak_index_mult"]]
            if "_pd_peak_2theta" in l_name:
                self.ttheta = [float(_1) for _1 in cif_loop["_pd_peak_2theta"]]
            if "_pd_peak_intensity_up" in l_name:
                self.up = [float(_1) for _1 in cif_loop["_pd_peak_intensity_up"]]
            if "_pd_peak_intensity_down" in l_name:
                self.down = [float(_1) for _1 in cif_loop["_pd_peak_intensity_down"]]
            if "_pd_peak_width_2theta" in l_name:
                self.width_2theta = [float(_1) for _1 in cif_loop["_pd_peak_width_2theta"]]
        else:
            self.h, self.k, self.l, self.mult, self.ttheta, self.up, self.down, self.width_2theta = [], [], [], [], [], [], [], []
        return True

    @property
    def is_defined(self):
        cond = all([self.h is not None, self.k is not None, self.l is not None, self.mult is not None, 
                    self.ttheta is not None, self.up is not None, self.down is not None,
                    self.width_2theta is not None])
        return cond

    @property
    def is_variable(self):
        return False
    
    def get_variables(self):
        return []

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)



"""
define classes to describe AtomType
"""
__author__ = 'ikibalin'
__version__ = "2019_08_26"
import os
import numpy


from cryspy.f_common.cl_fitable import Fitable

class AtomType(object):
    """
    Data items in the ATOM_TYPE category record details about
    properties of the atoms that occupy the atom sites, such as the
    atomic scattering factors.
    
    Description in cif file:

    loop_
    _atom_type_symbol
    _atom_type_oxidation_number
    _atom_type_number_in_cell
    _atom_type_scat_dispersion_real
    _atom_type_scat_dispersion_imag
    _atom_type_scat_source
      C 0 72  .017  .009  International_Tables_Vol_IV_Table_2.2B
      H 0 100  0     0    International_Tables_Vol_IV_Table_2.2B
      O 0 12  .047  .032  International_Tables_Vol_IV_Table_2.2B
      N 0 4   .029  .018  International_Tables_Vol_IV_Table_2.2B

    reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Catom_type.html
    
    """    
    def __init__(self, symbol=[], scat_length_neutron=[]):
        super(AtomType, self).__init__()

        self.__atom_type_symbol = []
        self.__atom_type_scat_length_neutron = []

    @property
    def symbol(self):
        """
        """
        return tuple(self.__atom_type_symbol)
    @symbol.setter
    def symbol(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = str(x).strip()
            l_x_in.append(x_in)
        self.__atom_type_symbol = l_x_in
        len_x = len(l_x_in)
        len_1 = len(self.__atom_type_scat_length_neutron)
        if len_1 > len_x:
            self.__atom_type_scat_length_neutron = self.__atom_type_scat_length_neutron[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_type_scat_length_neutron") for hh in range(len_x-len_1)]
            self.__atom_type_scat_length_neutron.extend(l_fitable)


    @property
    def scat_length_neutron(self):
        """
        in 10**-12 cm 
        """
        return tuple(self.__atom_type_scat_length_neutron)
    @scat_length_neutron.setter
    def scat_length_neutron(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_type_symbol)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name="_atom_type_scat_length_neutron") for hh in range(len_1-len_x)])
        self.__atom_type_scat_length_neutron = l_fitable

    @property
    def b_scat(self):
        return self.scat_length_neutron
    @b_scat.setter
    def b_scat(self, l_x):
        self.scat_length_neutron = l_x

    def __repr__(self):
        ls_out = ["AtomType:"]
        ls_out.append(" symbol   scat_length_neutron")
        for _1, _2 in zip(self.symbol, self.scat_length_neutron):
            ls_out.append(" {:8} {:8}".format(_1, _2.print_with_sigma))
        return "\n".join(ls_out)

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    @property
    def is_defined(self):
        """
        Output: True if all started parameters are given
        """
        cond = (self.symbol != [])
        return cond

    @property
    def is_variable(self):
        res = (any([hh.refinement for hh in self.scat_length_neutron]))
        return res
    def get_variables(self):
        l_variable = []
        l_variable.extend([hh for hh in self.scat_length_neutron if hh.refinement])
        return l_variable

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_atom_type_symbol")
            ls_out.append("_atom_type_scat_length_neutron")
            for _1, _2 in zip(self.symbol, self.scat_length_neutron):
                ls_out.append("{:} {:}".format(_1, _2.print_with_sigma))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_atom_type")
        if flag:
            cif_loop = cif_global["_atom_type"]
            l_name = cif_loop.names
            if "_atom_type_symbol" in l_name:
                self.symbol = cif_loop["_atom_type_symbol"]

            if "_atom_type_scat_length_neutron" in l_name:
                l_val = cif_loop["_atom_type_scat_length_neutron"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_type_scat_length_neutron")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.scat_length_neutron = l_fitable
        else:
            self.symbol = []
        return True

    def _form_scat_length_neutron(self, atom_site):
        """
        Note: It is not correct!!! It should be changed if it is neaded
        """
        label = numpy.array(atom_site.label, dtype=str)
        label_magnetism = numpy.array(self.label, dtype=str)
        if not(set(label_magnetism).issubset(set(label))):
            self._show_message("Unknown 'aniso_label'")
            return False
        np_index = numpy.array([int(numpy.argwhere(label==hh)[0]) for hh in label_magnetism], dtype=int)
        scat_length_neutron_in = numpy.zeros(label.shape, dtype=float)
        scat_length_neutron = numpy.array(self.scat_length_neutron, dtype=float)
        scat_length_neutron_in[np_index] = scat_length_neutron
        return scat_length_neutron_in


"""
define classe PdProc for 1d powder diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_06"
import os
import numpy


from pycifstar import Global


class PdProc(object):
    """
    This section contains the diffraction data set after processing
    and application of correction terms. If the data set is
    reprocessed, this section may be replaced (with the addition of
    a new _pd_block_id entry).

    Example:

    loop_
    _pd_proc_2theta
    _pd_proc_2theta_corrected
    _pd_proc_d_spacing
    _pd_proc_intensity_up_net
    _pd_proc_intensity_down_net
    _pd_proc_intensity_up_total
    _pd_proc_intensity_down_total
    _pd_proc_intensity_bkg_calc
    _pd_proc_intensity_up
    _pd_proc_intensity_up_sigma
    _pd_proc_intensity_down
    _pd_proc_intensity_down_sigma
     4.00  3.9  11.2   400.00   317.00   460.00   377.00   60.00    465.80000   128.97000   301.88000   129.30000

    reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_pd.dic/Cpd_proc.html
    """
    def __init__(self, ttheta=[], ttheta_corrected=[], d=[], 
                up_net=[], down_net=[], up_total=[], down_total=[], bkg_calc=[], 
                up=[], down=[], up_sigma=[], down_sigma=[]):
        super(PdProc, self).__init__()
        self.__pd_proc_2theta = None
        self.__pd_proc_2theta_corrected = None
        self.__pd_proc_d_spacing = None
        self.__pd_proc_intensity_up_net = None
        self.__pd_proc_intensity_down_net = None
        self.__pd_proc_intensity_up_total = None
        self.__pd_proc_intensity_down_total = None
        self.__pd_proc_intensity_bkg_calc = None
        self.__pd_proc_intensity_up = None
        self.__pd_proc_intensity_up_sigma = None
        self.__pd_proc_intensity_down = None
        self.__pd_proc_intensity_down_sigma = None



        self.ttheta = ttheta
        self.ttheta_corrected = ttheta_corrected
        self.d = d
        self.up_net = up_net
        self.down_net = down_net
        self.up_total = up_total
        self.down_total = down_total
        self.bkg_calc = bkg_calc
        self.up = up
        self.up_sigma = up_sigma
        self.down = down
        self.down_sigma = down_sigma

    @property
    def ttheta(self):
        return self.__pd_proc_2theta
    @ttheta.setter
    def ttheta(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_2theta = np_x_in

    @property
    def ttheta_corrected(self):
        return self.__pd_proc_2theta_corrected
    @ttheta_corrected.setter
    def ttheta_corrected(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_2theta_corrected = np_x_in


    @property
    def d(self):
        return self.__pd_proc_d_spacing
    @d.setter
    def d(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_d_spacing = np_x_in

    @property
    def up_net(self):
        return self.__pd_proc_intensity_up_net
    @up_net.setter
    def up_net(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_intensity_up_net = np_x_in


    @property
    def down_net(self):
        return self.__pd_proc_intensity_down_net
    @down_net.setter
    def down_net(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_intensity_down_net = np_x_in


    @property
    def up_total(self):
        return self.__pd_proc_intensity_up_total
    @up_total.setter
    def up_total(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_intensity_up_total = np_x_in


    @property
    def down_total(self):
        return self.__pd_proc_intensity_down_total
    @down_total.setter
    def down_total(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_intensity_down_total = np_x_in


    @property
    def bkg_calc(self):
        return self.__pd_proc_intensity_bkg_calc
    @bkg_calc.setter
    def bkg_calc(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_intensity_bkg_calc = np_x_in


    @property
    def up(self):
        return self.__pd_proc_intensity_up
    @up.setter
    def up(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_intensity_up = np_x_in

    @property
    def up_sigma(self):
        return self.__pd_proc_intensity_up_sigma
    @up_sigma.setter
    def up_sigma(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_intensity_up_sigma = np_x_in

    @property
    def down(self):
        return self.__pd_proc_intensity_down
    @down.setter
    def down(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_intensity_down = np_x_in

    @property
    def down_sigma(self):
        return self.__pd_proc_intensity_down_sigma
    @down_sigma.setter
    def down_sigma(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_proc_intensity_down_sigma = np_x_in


            
    def __repr__(self):
        ls_out = ["PdProc:"]
        ls_out.append(" ttheta  ttheta_cor. up_total           up   down_total         down")
        for _1, _2, _3, _4, _5, _6 in zip(self.ttheta, self.ttheta_corrected, self.up_total, self.up, self.down_total, self.down):
            ls_out.append("{:7.2f} {:7.2f} {:12.5f} {:12.5f} {:12.5f} {:12.5f}".format(_1, _2, _3, _4, _5, _6))
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_pd_proc_2theta")
            ls_out.append("_pd_proc_2theta_corrected")
            #ls_out.append("_pd_proc_d_spacing")
            ls_out.append("_pd_proc_intensity_up_net")
            ls_out.append("_pd_proc_intensity_down_net")
            ls_out.append("_pd_proc_intensity_up_total")
            ls_out.append("_pd_proc_intensity_down_total")
            ls_out.append("_pd_proc_intensity_bkg_calc")
            ls_out.append("_pd_proc_intensity_up")
            ls_out.append("_pd_proc_intensity_up_sigma")
            ls_out.append("_pd_proc_intensity_down")
            ls_out.append("_pd_proc_intensity_down_sigma")
            #for _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11, _12 in zip(self.ttheta, self.ttheta_corrected, self.d, 
            #    self.up_net, self.down_net, self.up_total, self.down_total, self.bkg_calc, 
            #    self.up, self.up_sigma, self.down, self.down_sigma):
            #    ls_out.append("{:} {:} {:} {:} {:} {:} {:} {:} {:} {:} {:} {:}".format(_1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11, _12))
            for _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11 in zip(self.ttheta, self.ttheta_corrected, 
                self.up_net, self.down_net, self.up_total, self.down_total, self.bkg_calc, 
                self.up, self.up_sigma, self.down, self.down_sigma):
                ls_out.append("{:.3f} {:.3f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f}".format(_1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_pd_proc")
        if flag:
            cif_loop = cif_global["_pd_proc"]
            l_name = cif_loop.names
            if "_pd_proc_2theta" in l_name:
                self.ttheta = [float(_1) for _1 in cif_loop["_pd_proc_2theta"]]
            if "_pd_proc_2theta_corrected" in l_name:
                self.ttheta_corrected = [float(_1) for _1 in cif_loop["_pd_proc_2theta_corrected"]]
            if "_pd_proc_d_spacing" in l_name:
                self.d = [float(_1) for _1 in cif_loop["_pd_proc_d_spacing"]]
            if "_pd_proc_intensity_up_net" in l_name:
                self.up_net = [float(_1) for _1 in cif_loop["_pd_proc_intensity_up_net"]]
            if "_pd_proc_intensity_down_net" in l_name:
                self.down_net = [float(_1) for _1 in cif_loop["_pd_proc_intensity_down_net"]]
            if "_pd_proc_intensity_up_total" in l_name:
                self.up_total = [float(_1) for _1 in cif_loop["_pd_proc_intensity_up_total"]]
            if "_pd_proc_intensity_down" in l_name:
                self.down_total = [float(_1) for _1 in cif_loop["_pd_proc_intensity_down_total"]]
            if "_pd_proc_intensity_bkg_calc" in l_name:
                self.bkg_calc = [float(_1) for _1 in cif_loop["_pd_proc_intensity_bkg_calc"]]
            if "_pd_proc_intensity_up" in l_name:
                self.up = [float(_1) for _1 in cif_loop["_pd_proc_intensity_up"]]
            if "_pd_proc_intensity_up_sigma" in l_name:
                self.up_sigma = [float(_1) for _1 in cif_loop["_pd_proc_intensity_up_sigma"]]
            if "_pd_proc_intensity_down" in l_name:
                self.down = [float(_1) for _1 in cif_loop["_pd_proc_intensity_down"]]
            if "_pd_proc_intensity_down_sigma" in l_name:
                self.down_sigma = [float(_1) for _1 in cif_loop["_pd_proc_intensity_down_sigma"]]
        else:
            self.ttheta, self.ttheta_corrected, self.d = [], [], []
            self.up_net, self.down_net, self.up_total, self.down_total, self.bkg_calc = [], [], [], [], []
            self.up, self.up_sigma, self.down, self.down_sigma = [], [], [], []
        return True

    @property
    def is_defined(self):
        cond = all([self.ttheta is not None, self.ttheta_corrected is not None, self.d is not None, 
                    self.up_net is not None, self.down_net is not None, self.up_total is not None, self.down_total is not None,
                    self.bkg_calc is not None,
                    self.up is not None, self.up_sigma is not None, self.down is not None, self.down_sigma is not None])
        return cond

    @property
    def is_variable(self):
        return False
    
    def get_variables(self):
        return []

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)



"""
define classes to describe Crystal
"""
__author__ = 'ikibalin'
__version__ = "2019_09_09"
import os
import numpy
import copy

from pycifstar import Data
from cryspy.f_common.cl_fitable import Fitable
from .cl_space_group import SpaceGroup
from .cl_cell import Cell
from .cl_atom_site import AtomSite
from .cl_atom_site_aniso import AtomSiteAniso
from .cl_atom_site_magnetism import AtomSiteMagnetism
from .cl_atom_site_magnetism_aniso import AtomSiteMagnetismAniso

from .cl_magnetism import calc_mRmCmRT

from .cl_refln import Refln


class Crystal(object):
    """
    Data items in the CRYSTAL category record details about
    crystal structure.
    
    Description in cif file:

    data_Fe3O4                                # object Crystal with label 'Fe3O4'
    _cell_angle_alpha 90.0                    # object Cell
    _cell_angle_beta 90.0
    _cell_angle_gamma 90.0
    _cell_length_a 8.56212()
    _cell_length_b 8.56212
    _cell_length_c 8.56212

    _space_group_it_coordinate_system_code 2  # object SpaceGroup
    _space_group_name_H-M_alt "F d -3 m"
    _space_group_IT_number    232

    loop_                                     # object AtomSite
    _atom_site_adp_type
    _atom_site_B_iso_or_equiv
    _atom_site_fract_x
    _atom_site_fract_y
    _atom_site_fract_z
    _atom_site_label
    _atom_site_occupancy
    _atom_site_type_symbol
     uani 0.0 0.125 0.125 0.125 Fe3A 1.0 Fe3+
     uani 0.0 0.5 0.5 0.5 Fe3B 1.0 Fe3+
     uiso 0.0 0.25521 0.25521 0.25521 O1 1.0 O2-

    loop_                                     # object AtomType (optional)
    _atom_type_scat_length_neutron
    _atom_type_symbol
      0.945 Fe3+
     0.5803 O2-

    loop_                                     # object AtomSiteAniso (optional)
    _atom_site_aniso_U_11
    _atom_site_aniso_U_12
    _atom_site_aniso_U_13
    _atom_site_aniso_U_22
    _atom_site_aniso_U_23
    _atom_site_aniso_U_33
    _atom_site_aniso_label
     0.0 0.0 0.0 0.0 0.0 0.0 Fe3A
     0.0 0.0 0.0 0.0 0.0 0.0 Fe3B

    loop_                                     # object AtomSiteMagnetism (optional)
    _atom_site_magnetism_label
    _atom_site_magnetism_lande
    _atom_site_magnetism_kappa
    Fe3A 2.0 1.0()
    Fe3B 2.0() 1.0

    loop_                                     # object AtomSiteMagnetismAniso (optional)
    _atom_site_magnetism_aniso_label
    _atom_site_magnetism_aniso_chi_type
    _atom_site_magnetism_aniso_chi_11
    _atom_site_magnetism_aniso_chi_12
    _atom_site_magnetism_aniso_chi_13
    _atom_site_magnetism_aniso_chi_22
    _atom_site_magnetism_aniso_chi_23
    _atom_site_magnetism_aniso_chi_33
     Fe3A cani -3.468(74) 0.0 0.0 -3.468 0.0 -3.468
     Fe3B cani 3.041      0.0 0.0  3.041 0.0  3.041
    """    
    def __init__(self, label='',  cell=None, space_group=None, atom_type=None,
                       atom_site=None, atom_site_aniso=None, 
                       atom_site_magnetism=None, atom_site_magnetism_aniso=None):
        super(Crystal, self).__init__()

        self.__label = ""
        self.__cell = None
        self.__space_group = None
        self.__atom_type = None
        self.__atom_site = None
        self.__atom_site_aniso = None
        self.__atom_site_magnetism = None
        self.__atom_site_magnetism_aniso = None

        self.label = label
        self.cell = cell
        self.space_group = space_group
        self.atom_type = atom_type
        self.atom_site = atom_site
        self.atom_site_aniso = atom_site_aniso
        self.atom_site_magnetism = atom_site_magnetism
        self.atom_site_magnetism_aniso = atom_site_magnetism_aniso
        
    def __repr__(self):
        ls_out = ["Crystal:"]
        ls_out.append("label: "+self.label+"\n")
        if self.cell is not None:
            ls_out.append(str(self.cell)+"\n")
        if self.space_group is not None:
            ls_out.append(str(self.space_group)+"\n")
        if self.atom_type is not None:
            ls_out.append(str(self.atom_type)+"\n")
        if self.atom_site is not None:
            ls_out.append(str(self.atom_site)+"\n")
        if self.atom_site_aniso is not None:
            ls_out.append(str(self.atom_site_aniso)+"\n")
        if self.atom_site_magnetism is not None:
            ls_out.append(str(self.atom_site_magnetism)+"\n")
        if self.atom_site_magnetism_aniso is not None:
            ls_out.append(str(self.atom_site_magnetism_aniso)+"\n")
        return "\n".join(ls_out)

    @property
    def label(self):
        """
        The label is a unique identifier for a particular crystal. 

        Type: char
        """
        return self.__label
    @label.setter
    def label(self, x):
        if x is None:
            x_in = ""
        else:
            x_in = str(x).strip()
        self.__label = x_in

    @property
    def cell(self):
        """
        Data items in the CELL category record details about the
        crystallographic cell parameters and their measurement.

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Ccell.html
        """
        return self.__cell
    @cell.setter
    def cell(self, x):
        if isinstance(x, Cell):
            x_in = x
        elif isinstance(x, str):
            x_in = Cell()
            flag = x_in.from_cif(x)
            if not(flag):
                self._show_message("A induced string can not be converted to Cell")
                x_in = None
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("A type of induced element is not recognized to convert it into Cell")
        self.__cell = x_in

    @property
    def space_group(self):
        """
        Contains all the data items that refer to the space group as a
        whole, such as its name or crystal system. They may be looped,
        for example, in a list of space groups and their properties.

        Only a subset of the SPACE_GROUP category items appear in the
        core dictionary.  The remainder are found in the symmetry CIF
        dictionary.

        Space-group types are identified by their number as given in
        International Tables for Crystallography Vol. A. Specific
        settings of the space groups can be identified either by their
        Hall symbol or by specifying their symmetry operations.

        The commonly used Hermann-Mauguin symbol determines the
        space-group type uniquely but several different Hermann-Mauguin
        symbols may refer to the same space-group type. A
        Hermann-Mauguin symbol contains information on the choice of
        the basis, but not on the choice of origin.  Different formats
        for the Hermann-Mauguin symbol are found in the symmetry CIF
        dictionary.

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Cspace_group.html
        """
        return self.__space_group
    @space_group.setter
    def space_group(self, x):
        if isinstance(x, SpaceGroup):
            x_in = x
        elif isinstance(x, str):
            x_in = SpaceGroup()
            flag = x_in.from_cif(x)
            if not(flag):
                self._show_message("A induced string can not be converted to SpaceGroup")
                x_in = None
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("A type of induced element is not recognized to convert it into SpaceGroup")
        self.__space_group = x_in

    @property
    def atom_type(self):
        """
        Data items in the ATOM_TYPE category record details about
        properties of the atoms that occupy the atom sites, such as the
        atomic scattering factors.

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Catom_type.html
        """
        return self.__atom_type
    @atom_type.setter
    def atom_type(self, x):
        """
        if isinstance(x, AtomType):
            x_in = x
        elif isinstance(x, str):
            x_in = AtomType()
            flag = x_in.from_cif(x)
            if not(flag):
                self._show_message("A induced string can not be converted to AtomType")
                x_in = None
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("A type of induced element is not recognized to convert it into AtomType")
        self.__atom_type = x_in
        """
        self.__atom_type = None #class AtomType is still not introduced


    @property
    def atom_site(self):
        """
        Data items in the ATOM_SITE category record details about
        the atom sites in a crystal structure, such as the positional
        coordinates.

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Catom_site.html
        """
        return self.__atom_site
    @atom_site.setter
    def atom_site(self, x):
        if isinstance(x, AtomSite):
            x_in = x
        elif isinstance(x, str):
            x_in = AtomSite()
            flag = x_in.from_cif(x)
            if not(flag):
                self._show_message("A induced string can not be converted to AtomSite")
                x_in = None
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("A type of induced element is not recognized to convert it into AtomSite")
        self.__atom_site = x_in

    @property
    def atom_site_aniso(self):
        """
        Data items in the ATOM_SITE_ANISO category record details about
        the atom sites in a crystal structure, such as atomic displacement 
        parameters.

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Catom_site.html
        """
        return self.__atom_site_aniso
    @atom_site_aniso.setter
    def atom_site_aniso(self, x):
        if isinstance(x, AtomSiteAniso):
            x_in = x
        elif isinstance(x, str):
            x_in = AtomSiteAniso()
            flag = x_in.from_cif(x)
            if not(flag):
                self._show_message("A induced string can not be converted to AtomSiteAniso")
                x_in = None
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("A type of induced element is not recognized to convert it into AtomSiteAniso")
        self.__atom_site_aniso = x_in


    @property
    def atom_site_magnetism(self):
        """
        Data items in the ATOM_SITE_MAGNETISM category record details about
        the magnetic parameters.

        """
        return self.__atom_site_magnetism
    @atom_site_magnetism.setter
    def atom_site_magnetism(self, x):
        if isinstance(x, AtomSiteMagnetism):
            x_in = x
        elif isinstance(x, str):
            x_in = AtomSiteMagnetism()
            flag = x_in.from_cif(x)
            if not(flag):
                self._show_message("A induced string can not be converted to AtomSiteMagnetism")
                x_in = None
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("A type of induced element is not recognized to convert it into AtomSiteMagnetism")
        self.__atom_site_magnetism = x_in


    @property
    def atom_site_magnetism_aniso(self):
        """
        Data items in the ATOM_SITE_MAGNETISM_ANISO category record details about
        the susceptibility tensor.
        """
        return self.__atom_site_magnetism_aniso
    @atom_site_magnetism_aniso.setter
    def atom_site_magnetism_aniso(self, x):
        if isinstance(x, AtomSiteMagnetismAniso):
            x_in = x
        elif isinstance(x, str):
            x_in = AtomSiteMagnetismAniso()
            flag = x_in.from_cif(x)
            if not(flag):
                self._show_message("A induced string can not be converted to AtomSiteMagnetismAniso")
                x_in = None
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("A type of induced element is not recognized to convert it into AtomSiteMagnetismAniso")
        self.__atom_site_magnetism_aniso = x_in

    @property
    def magnetism_aniso(self):
        return self.atom_site_magnetism_aniso
    @magnetism_aniso.setter
    def magnetism_aniso(self, x):
        self.atom_site_magnetism_aniso = x


    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    @property
    def is_defined(self):
        """
        Output: True if all started parameters are given
        """
        cond = any([self.cell is not None, self.space_group is not None, self.atom_site is not None])
        return cond
    @property
    def is_variable(self):
        flag_1, flag_2, flag_3, flag_4, flag_5, flag_6, flag_7 = False, False, False, False, False, False, False
        if self.cell is not None: flag_1 = self.cell.is_variable
        #if self.space_group is not None: flag_2 = self.space_group.is_variable
        if self.atom_type is not None: flag_3 = self.atom_type.is_variable
        if self.atom_site is not None: flag_4 = self.atom_site.is_variable
        if self.atom_site_aniso is not None: flag_5 = self.atom_site_aniso.is_variable
        if self.atom_site_magnetism is not None: flag_6 = self.atom_site_magnetism.is_variable
        if self.atom_site_magnetism_aniso is not None: flag_7 = self.atom_site_magnetism_aniso.is_variable
        res = any([flag_1, flag_2, flag_3, flag_4, flag_5, flag_6, flag_7])
        return res


    def get_variables(self):
        l_val_1, l_val_2, l_val_3, l_val_4, l_val_5, l_val_6, l_val_7 = [], [], [], [], [], [], []
        if self.cell is not None: l_val_1 = self.cell.get_variables()
        #if self.space_group is not None: l_val_2 = self.space_group.get_variables()
        if self.atom_type is not None: l_val_3 = self.atom_type.get_variables()
        if self.atom_site is not None: l_val_4 = self.atom_site.get_variables()
        if self.atom_site_aniso is not None: l_val_5 = self.atom_site_aniso.get_variables()
        if self.atom_site_magnetism is not None: l_val_6 = self.atom_site_magnetism.get_variables()
        if self.atom_site_magnetism_aniso is not None: l_val_7 = self.atom_site_magnetism_aniso.get_variables()

        l_variable = []
        l_variable.extend(l_val_1)
        l_variable.extend(l_val_2)
        l_variable.extend(l_val_3)
        l_variable.extend(l_val_4)
        l_variable.extend(l_val_5)
        l_variable.extend(l_val_6)
        l_variable.extend(l_val_7)
        return l_variable

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            str_1, str_2, str_3, str_4, str_5, str_6, str_7 = "", "", "", "", "", "", ""
            ls_out.append("data_{:}".format(self.label))
            if self.cell is not None: str_1 = self.cell.to_cif + "\n"
            if self.space_group is not None: str_2 = self.space_group.to_cif + "\n"
            if self.atom_type is not None: str_3 = self.atom_type.to_cif + "\n"
            if self.atom_site is not None: str_4 = self.atom_site.to_cif + "\n"
            if self.atom_site_aniso is not None: str_5 = self.atom_site_aniso.to_cif + "\n"
            if self.atom_site_magnetism is not None: str_6 = self.atom_site_magnetism.to_cif + "\n"
            if self.atom_site_magnetism_aniso is not None: str_7 = self.atom_site_magnetism_aniso.to_cif + "\n"
            ls_out.extend([str_1, str_2, str_3, str_4, str_5, str_6, str_7])
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_data = Data()
        flag = cif_data.take_from_string(string)
        if not flag:
            return False
        self.label = cif_data.name
        cif_values = cif_data.items
        if cif_values is not None:
            if cif_values.is_prefix("cell"):
                self.cell = str(cif_values)
            if cif_values.is_prefix("space_group"):
                self.space_group = str(cif_values)
        if cif_data.is_prefix("atom_type"): self.atom_type = str(cif_data["atom_type"])
        if cif_data.is_prefix("atom_site"): self.atom_site = str(cif_data["atom_site"])
        if cif_data.is_prefix("atom_site_aniso"): self.atom_site_aniso = str(cif_data["atom_site_aniso"])
        if cif_data.is_prefix("atom_site_magnetism"): self.atom_site_magnetism = str(cif_data["atom_site_magnetism"])
        if cif_data.is_prefix("atom_site_magnetism_aniso"): self.atom_site_magnetism_aniso = str(cif_data["atom_site_magnetism_aniso"])
        return True

    def calc_refln(self, h=0, k=0, l=0):
        """
        calculate refln object
        """
        try:
            for _1, _2, _3 in zip(h, k, l):
                pass
            h_in = numpy.array(h, dtype=int)
            k_in = numpy.array(k, dtype=int)
            l_in = numpy.array(l, dtype=int)
        except:
            h_in = numpy.array([h], dtype=int)
            k_in = numpy.array([k], dtype=int)
            l_in = numpy.array([l], dtype=int)
        refln = self.calc_sf(h_in, k_in, l_in)
        return refln

    def calc_sf(self, h, k, l):
        """
        calculate nuclear structure factor and components of structure factor tensor
        """

        space_group = self.space_group
        cell = self.cell
        atom_site = self.atom_site
        atom_site_aniso = self.atom_site_aniso
        atom_site_magnetism = self.atom_site_magnetism
        atom_site_magnetism_aniso = self.atom_site_magnetism_aniso


        occupancy = numpy.array(atom_site.occupancy, dtype=float)
        scat_length_neutron = numpy.array(atom_site.scat_length_neutron, dtype=complex)


        fract = atom_site._form_fract()
        flag_adp = atom_site_aniso is not None
        flag_magnetism = atom_site_magnetism_aniso is not None
        if flag_adp:
            adp = atom_site_aniso._form_adp(atom_site)
        if flag_magnetism:
            magnetism = atom_site_magnetism_aniso._form_magnetism(atom_site, atom_site_magnetism)


        x, y, z = fract.x, fract.y, fract.z

        atom_multiplicity = space_group.calc_atom_mult(x, y, z)
        atom_site.multiplicity = atom_multiplicity
        occ_mult = occupancy*atom_multiplicity 
        

        phase_3d = fract.calc_phase(space_group, h, k, l)#3d object
        if flag_adp:
            dwf_3d = adp.calc_dwf(space_group, cell, h, k, l)
        else:
            dwf_3d = numpy.ones(phase_3d.shape, dtype=float)

        hh = phase_3d*dwf_3d

        if flag_magnetism:
            ff_11, ff_12, ff_13, ff_21, ff_22, ff_23, ff_31, ff_32, ff_33 = \
                   magnetism.calc_form_factor_tensor(space_group, cell, h, k, l)
        else:
            np_zeros = numpy.zeros(phase_3d.shape, dtype=float)
            ff_11, ff_12, ff_13 = np_zeros, np_zeros, np_zeros
            ff_21, ff_22, ff_23 = np_zeros, np_zeros, np_zeros 
            ff_31, ff_32, ff_33 = np_zeros, np_zeros, np_zeros


        phase_2d = hh.sum(axis=2)

        ft_11 = (ff_11*hh).sum(axis=2)
        ft_12 = (ff_12*hh).sum(axis=2)
        ft_13 = (ff_13*hh).sum(axis=2)
        ft_21 = (ff_21*hh).sum(axis=2)
        ft_22 = (ff_22*hh).sum(axis=2)
        ft_23 = (ff_23*hh).sum(axis=2)
        ft_31 = (ff_31*hh).sum(axis=2)
        ft_32 = (ff_32*hh).sum(axis=2)
        ft_33 = (ff_33*hh).sum(axis=2)
        
        b_scat_2d = numpy.meshgrid(h, scat_length_neutron, indexing="ij")[1]
        occ_mult_2d = numpy.meshgrid(h, occ_mult, indexing="ij")[1]
        
        l_el_symm = space_group.el_symm
        l_orig = space_group.orig
        centr = space_group.centr

        #calculation of nuclear structure factor        
        hh = phase_2d * b_scat_2d * occ_mult_2d
        f_hkl_as = hh.sum(axis=1)*1./len(l_el_symm)
        
        orig_x = [hh[0] for hh in l_orig]
        orig_y = [hh[1] for hh in l_orig]
        orig_z = [hh[2] for hh in l_orig]
        
        np_h, np_orig_x = numpy.meshgrid(h, orig_x, indexing = "ij")
        np_k, np_orig_y = numpy.meshgrid(k, orig_y, indexing = "ij")
        np_l, np_orig_z = numpy.meshgrid(l, orig_z, indexing = "ij")
        
        np_orig_as = numpy.exp(2*numpy.pi*1j*(np_h*np_orig_x+np_k*np_orig_y+np_l*np_orig_z))
        f_hkl_as = f_hkl_as*np_orig_as.sum(axis=1)*1./len(l_orig)

        if (centr):
            orig = space_group.p_centr
            f_nucl = 0.5*(f_hkl_as+f_hkl_as.conjugate()*numpy.exp(2.*2.*numpy.pi*1j* (h*orig[0]+k*orig[1]+l*orig[2])))
        else:
            f_nucl = f_hkl_as

        #calculation of structure factor tensor
        sft_as_11 = (ft_11 * occ_mult_2d).sum(axis=1)*1./len(l_el_symm)
        sft_as_12 = (ft_12 * occ_mult_2d).sum(axis=1)*1./len(l_el_symm)
        sft_as_13 = (ft_13 * occ_mult_2d).sum(axis=1)*1./len(l_el_symm)
        sft_as_21 = (ft_21 * occ_mult_2d).sum(axis=1)*1./len(l_el_symm)
        sft_as_22 = (ft_22 * occ_mult_2d).sum(axis=1)*1./len(l_el_symm)
        sft_as_23 = (ft_23 * occ_mult_2d).sum(axis=1)*1./len(l_el_symm)
        sft_as_31 = (ft_31 * occ_mult_2d).sum(axis=1)*1./len(l_el_symm)
        sft_as_32 = (ft_32 * occ_mult_2d).sum(axis=1)*1./len(l_el_symm)
        sft_as_33 = (ft_33 * occ_mult_2d).sum(axis=1)*1./len(l_el_symm)

        sft_as_11 = sft_as_11 * np_orig_as.sum(axis=1)*1./len(l_orig)
        sft_as_12 = sft_as_12 * np_orig_as.sum(axis=1)*1./len(l_orig)
        sft_as_13 = sft_as_13 * np_orig_as.sum(axis=1)*1./len(l_orig)
        sft_as_21 = sft_as_21 * np_orig_as.sum(axis=1)*1./len(l_orig)
        sft_as_22 = sft_as_22 * np_orig_as.sum(axis=1)*1./len(l_orig)
        sft_as_23 = sft_as_23 * np_orig_as.sum(axis=1)*1./len(l_orig)
        sft_as_31 = sft_as_31 * np_orig_as.sum(axis=1)*1./len(l_orig)
        sft_as_32 = sft_as_32 * np_orig_as.sum(axis=1)*1./len(l_orig)
        sft_as_33 = sft_as_33 * np_orig_as.sum(axis=1)*1./len(l_orig)
    
        if (centr):
            orig = space_group.p_centr
            hh = numpy.exp(2.*2.*numpy.pi*1j* (h*orig[0]+k*orig[1]+l*orig[2]))
            sft_11 = 0.5*(sft_as_11+sft_as_11.conjugate()*hh)
            sft_12 = 0.5*(sft_as_12+sft_as_12.conjugate()*hh)
            sft_13 = 0.5*(sft_as_13+sft_as_13.conjugate()*hh)
            sft_21 = 0.5*(sft_as_21+sft_as_21.conjugate()*hh)
            sft_22 = 0.5*(sft_as_22+sft_as_22.conjugate()*hh)
            sft_23 = 0.5*(sft_as_23+sft_as_23.conjugate()*hh)
            sft_31 = 0.5*(sft_as_31+sft_as_31.conjugate()*hh)
            sft_32 = 0.5*(sft_as_32+sft_as_32.conjugate()*hh)
            sft_33 = 0.5*(sft_as_33+sft_as_33.conjugate()*hh)
        else:
            sft_11, sft_12, sft_13 = sft_as_11, sft_as_12, sft_as_13
            sft_21, sft_22, sft_23 = sft_as_21, sft_as_22, sft_as_23
            sft_31, sft_32, sft_33 = sft_as_31, sft_as_32, sft_as_33            

        #sft_ij form the structure factor tensor in local coordinate system (ia, ib, ic)
        #chi in 10-12 cm; chim in muB (it is why here 0.2695)
        s_11, s_12, s_13, s_21, s_22, s_23, s_31, s_32, s_33 = self._orto_matrix(
                cell,
                sft_11*0.2695, sft_12*0.2695, sft_13*0.2695, 
                sft_21*0.2695, sft_22*0.2695, sft_23*0.2695, 
                sft_31*0.2695, sft_32*0.2695, sft_33*0.2695)

        refln = Refln()
        refln.h, refln.k, refln.l = copy.deepcopy(h), copy.deepcopy(k), copy.deepcopy(l)
        refln.f_nucl = f_nucl
        refln.sft_11, refln.sft_12, refln.sft_13 = s_11, s_12, s_13
        refln.sft_21, refln.sft_22, refln.sft_23 = s_21, s_22, s_23
        refln.sft_31, refln.sft_32, refln.sft_33 = s_31, s_32, s_33
        return refln


    def _orto_matrix(self, cell, l_11, l_12, l_13, l_21, l_22, l_23, l_31, 
                     l_32, l_33):
        """
        rewrite matrix l_ij defined in coordinate (ia, ib, ic) to matrix s_ij, 
        which is denined in Chartesian coordinate system, such as:
        x||ia, y in blane (ia, ib), z perpendicular to that plane.
        ...
        
        ...
        representation of chi in crystallographic coordinate system defined as x||a*, z||c, y= [z x] (right handed)
        expressions are taken from international tables
        matrix_ib is inversed matrix B
        ia, ib, ic is inversed unit cell parameters (it can be estimated from matrix matrix_ib)

        X = B x, x = iB X
        xT*CHI*x = XT iBT CHI iB X
    
        output chiLOC = iBT CHI iB
        """
        m_ib_norm = cell.m_ib_norm
        m_ibt_norm = m_ib_norm.transpose()
        
        r11, r12, r13 = m_ibt_norm[0, 0], m_ibt_norm[0, 1], m_ibt_norm[0, 2]
        r21, r22, r23 = m_ibt_norm[1, 0], m_ibt_norm[1, 1], m_ibt_norm[1, 2]
        r31, r32, r33 = m_ibt_norm[2, 0], m_ibt_norm[2, 1], m_ibt_norm[2, 2]
        
        s_11, s_12, s_13, s_21, s_22, s_23, s_31, s_32, s_33 = calc_mRmCmRT(
                r11, r12, r13, r21, r22, r23, r31, r32, r33,
                l_11, l_12, l_13, l_21, l_22, l_23, l_31, l_32, l_33)        

        return s_11, s_12, s_13, s_21, s_22, s_23, s_31, s_32, s_33

    def apply_constraint(self):
        space_group = self.space_group
        cell = self.cell
        cell.apply_constraint()
        atom_site = self.atom_site
        atom_site_aniso = self.atom_site_aniso
        if atom_site_aniso is not None:
            atom_site_aniso.apply_space_group_constraint(atom_site, space_group)
        atom_site_magnetism_aniso = self.atom_site_magnetism_aniso
        if atom_site_magnetism_aniso is not None:
            atom_site_magnetism_aniso.apply_chi_iso_constraint(cell)
            atom_site_magnetism_aniso.apply_space_group_constraint(atom_site, space_group)
    
    def calc_hkl(self, sthol_min, sthovl_max):
        cell = self.cell
        space_group = self.space_group
        res = cell.calc_hkl(space_group, sthol_min, sthovl_max)
        return res

    def calc_hkl_in_range(self, sthol_min, sthovl_max):
        cell = self.cell
        res = cell.calc_hkl_in_range(sthol_min, sthovl_max)
        return res



"""
define classe DiffrnRefln which describes the single diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_10"
import os
import numpy


from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable


class DiffrnRefln(object):
    """
    Data items in the DIFFRN_REFLN category record details about
    the intensities measured in the diffraction experiment.

    The DIFFRN_REFLN data items refer to individual intensity
    measurements and must be included in looped lists.

    Example:

    loop_
    _diffrn_refln_index_h
    _diffrn_refln_index_k
    _diffrn_refln_index_l
    _diffrn_refln_fr
    _diffrn_refln_fr_sigma
        0    0    8   0.64545   0.01329 
        2    0    6   1.75682   0.0454  
    
    reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Cdiffrn_refln.html
    """
    def __init__(self, h=[], k=[], l=[], fr=[], fr_sigma=[]
                 ):
        super(DiffrnRefln, self).__init__()
        self.__diffrn_refln_index_h = None
        self.__diffrn_refln_index_k = None
        self.__diffrn_refln_index_l = None
        self.__diffrn_refln_fr = None
        self.__diffrn_refln_fr_sigma = None
        self.h = h
        self.k = k
        self.l = l
        self.fr = fr
        self.fr_sigma = fr_sigma

        self.__diffrn_refln_fr_calc = None
        self.__diffrn_refln_intensity_up_calc = None
        self.__diffrn_refln_intensity_down_calc = None

    @property
    def h(self):
        return self.__diffrn_refln_index_h
    @h.setter
    def h(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, int):
                x_in = x
            else:
                x_in = int(round(x))
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__diffrn_refln_index_h = np_x_in

    @property
    def k(self):
        return self.__diffrn_refln_index_k
    @k.setter
    def k(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, int):
                x_in = x
            else:
                x_in = int(round(x))
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__diffrn_refln_index_k = np_x_in

    @property
    def l(self):
        return self.__diffrn_refln_index_l
    @l.setter
    def l(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, int):
                x_in = x
            else:
                x_in = int(round(x))
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__diffrn_refln_index_l = np_x_in

    @property
    def fr(self):
        return self.__diffrn_refln_fr
    @fr.setter
    def fr(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__diffrn_refln_fr = np_x_in


    @property
    def fr_sigma(self):
        return self.__diffrn_refln_fr_sigma
    @fr_sigma.setter
    def fr_sigma(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__diffrn_refln_fr_sigma = np_x_in


    @property
    def fr_calc(self):
        return self.__diffrn_refln_fr_calc
    @fr_calc.setter
    def fr_calc(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        if np_x_in.size != self.h.size:
            np_x_in = None
            self._show_message("Size of  fr_calc is different from size of hkl")
        self.__diffrn_refln_fr_calc = np_x_in

    @property
    def intensity_up_calc(self):
        return self.__diffrn_refln_intensity_up_calc
    @intensity_up_calc.setter
    def intensity_up_calc(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        if np_x_in.size != self.h.size:
            np_x_in = None
            self._show_message("Size of  intensity_up_calc is different from size of hkl")
        self.__diffrn_refln_intensity_up_calc = np_x_in

    @property
    def intensity_down_calc(self):
        return self.__diffrn_refln_intensity_down_calc
    @intensity_down_calc.setter
    def intensity_down_calc(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        if np_x_in.size != self.h.size:
            np_x_in = None
            self._show_message("Size of  intensity_down_calc is different from size of hkl")
        self.__diffrn_refln_intensity_down_calc = np_x_in
            
    def __repr__(self):
        ls_out = ["DiffrnRefln"]
        ls_out.append(str(self))
        return "\n".join(ls_out)

    def __str__(self):
        ls_out = []
        ls_out.append(" number of reflections: {:}".format(self.fr.size))
        ls_out.append("     min  max\n h  {:4} {:4}\n k  {:4} {:4}\n l  {:4} {:4}".format(
            self.h.min(), self.h.max(), self.k.min(), self.k.max(), self.l.min(), self.l.max()
        ))
        flag_fr_calc = self.fr_calc is not None
        if flag_fr_calc:
            ls_out.append("     h     k     l        fr  fr_sigma   fr_calc")
            for _1, _2, _3, _4, _5, _6 in zip(self.h, self.k, self.l, self.fr, self.fr_sigma, self.fr_calc):
                ls_out.append(" {:5} {:5} {:5} {:9.5} {:9.5} {:9.5}".format(_1, _2, _3, _4, _5, _6))
        else:
            ls_out.append("     h     k     l        fr  fr_sigma")
            for _1, _2, _3, _4, _5 in zip(self.h, self.k, self.l, self.fr, self.fr_sigma):
                ls_out.append(" {:5} {:5} {:5} {:9.5} {:9.5}".format(_1, _2, _3, _4, _5))

        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        flag_fr_calc = self.fr_calc is not None
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_diffrn_refln_index_h")
            ls_out.append("_diffrn_refln_index_k")
            ls_out.append("_diffrn_refln_index_l")
            ls_out.append("_diffrn_refln_fr")
            ls_out.append("_diffrn_refln_fr_sigma")
            if flag_fr_calc:
                ls_out.append("_diffrn_refln_fr_calc")
                for _1, _2, _3, _4, _5, _6 in zip(self.h, self.k, self.l, self.fr, self.fr_sigma, self.fr_calc):
                    ls_out.append("{:} {:} {:} {:} {:} {:}".format(_1, _2, _3, _4, _5, _6))
            else:
                for _1, _2, _3, _4, _5 in zip(self.h, self.k, self.l, self.fr, self.fr_sigma):
                    ls_out.append("{:} {:} {:} {:} {:}".format(_1, _2, _3, _4, _5))

        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_diffrn_refln")
        if flag:
            cif_loop = cif_global["_diffrn_refln"]
            l_name = cif_loop.names
            if "_diffrn_refln_index_h" in l_name:
                self.h = [int(_1) for _1 in cif_loop["_diffrn_refln_index_h"]]
            if "_diffrn_refln_index_k" in l_name:
                self.k = [int(_1) for _1 in cif_loop["_diffrn_refln_index_k"]]
            if "_diffrn_refln_index_l" in l_name:
                self.l = [int(_1) for _1 in cif_loop["_diffrn_refln_index_l"]]
            if "_diffrn_refln_fr" in l_name:
                self.fr = [float(_1) for _1 in cif_loop["_diffrn_refln_fr"]]
            if "_diffrn_refln_fr_sigma" in l_name:
                self.fr_sigma = [float(_1) for _1 in cif_loop["_diffrn_refln_fr_sigma"]]
        else:
            self.h, self.k, self.l, self.fr, self.fr_sigma = [], [], [], [], []
        return True

    @property
    def is_defined(self):
        cond = all([self.h is not None, self.k is not None, self.l is not None, self.fr is not None, self.fr_sigma is not None])
        return cond

    @property
    def is_variable(self):
        return False
    
    def get_variables(self):
        return []

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)
    def print_agreement_factor_exp(self):
        l_chi_sq_exp, l_ag_f_exp = [], []
        l_hkl = [(int(_1), int(_2), int(_3))for _1, _2, _3 in zip(self.h, self.k, self.l)]
        for _hkl, _fr_1, _fr_sigma_1 in zip(l_hkl, self.fr, self.fr_sigma):
            _mhkl = (-1*_hkl[0], -1*_hkl[1], -1*_hkl[2])
            if _mhkl in l_hkl:
                ind_mhkl = l_hkl.index(_mhkl)
                _fr_2, _fr_sigma_2 = self.fr[ind_mhkl], self.fr_sigma[ind_mhkl]
                _fr_sigma = 1./(_fr_sigma_1**(-2)+_fr_sigma_2**(-2))**0.5
                _fr_average = (_fr_1*_fr_sigma_1**(-2)+_fr_2*_fr_sigma_2**(-2))*_fr_sigma**2
                delta_fr = abs(_fr_1-_fr_average)
                chi_sq_exp = (delta_fr/_fr_sigma_1)**2
                l_chi_sq_exp.append(chi_sq_exp)
                ag_f_exp = abs((_fr_1-_fr_average)/(_fr_1-1.))
                l_ag_f_exp.append(ag_f_exp)
                #print("hkl: {:4} {:4} {:4}".format(_hkl[0], _hkl[1], _hkl[2]))
                #print("chi_sq_exp: {:.3f} ".format(chi_sq_exp))
                #print("ag_f_exp: {:.3f} ".format(ag_f_exp))
        ls_out = []
        n_friedel = len(l_chi_sq_exp)
        ls_out.append("number of Friedel reflections is {:}".format(n_friedel))
        if n_friedel != 0:
            ls_out.append("agreement factor_exp/n is {:.3f}".format(sum(l_ag_f_exp)/n_friedel))
            ls_out.append("chi_sq_exp/n is {:.3f}".format(sum(l_chi_sq_exp)/n_friedel))
        return "\n".join(ls_out) 




"""
internal class to calculate phases
"""
__author__ = 'ikibalin'
__version__ = "2019_09_02"
import os
import numpy


class Fract(object):
    """
    Fract of atom_sites in unit cell.
    """
    def __init__(self, x=numpy.array([0], dtype=float),
                       y=numpy.array([0], dtype=float),
                       z=numpy.array([0], dtype=float)):
        super(Fract, self).__init__()
        self.__atom_site_fract_x = None
        self.__atom_site_fract_y = None
        self.__atom_site_fract_z = None
        self.x = x 
        self.y = y
        self.z = z 

    def _trans_to_float_array(self, x):
        if isinstance(x, numpy.ndarray):
            x_out = x.astype(float)
        else:
            x_out = numpy.array([x], dtype=float)
        res = numpy.mod(x_out, 1.)
        return res

    @property
    def x(self):
        """

        reference:
        """
        return self.__atom_site_fract_x
    @x.setter
    def x(self, x):
        self.__atom_site_fract_x = self._trans_to_float_array(x)
    @property
    def y(self):
        """

        reference:
        """
        return self.__atom_site_fract_y
    @y.setter
    def y(self, x):
        self.__atom_site_fract_y = self._trans_to_float_array(x)
    @property
    def z(self):
        """

        reference:
        """
        return self.__atom_site_fract_z
    @z.setter
    def z(self, x):
        self.__atom_site_fract_z = self._trans_to_float_array(x)


    def __repr__(self):
        ls_out = ["Fract:\n       x        y        z"]
        ls_out.extend(["{:8.5f} {:8.5f} {:8.5f}".format(hh_1, hh_2, hh_3) \
                       for hh_1, hh_2, hh_3 in zip(self.x, self.y, self.z)])
        return "\n".join(ls_out)

    
    def calc_phase(self, space_group, h, k, l):
        """
        calculate phase: exp(-2 pi i * (h*x+k*y+l*z))
        r_11, r_22, r_33, r_12, r_13, r_23 are element of symmetry 
        """
        
        x, y, z = self.x, self.y, self.z

        r_11, r_12 = space_group.r_11, space_group.r_12
        r_13, r_21 = space_group.r_13, space_group.r_21
        r_22, r_23 = space_group.r_22, space_group.r_23
        r_31, r_32 = space_group.r_31, space_group.r_32
        r_33 = space_group.r_33
        b_1, b_2 = space_group.b_1, space_group.b_2
        b_3 = space_group.b_3
        
        np_h, np_x, np_r_11 = numpy.meshgrid(h, x, r_11, indexing="ij")
        np_k, np_y, np_r_22 = numpy.meshgrid(k, y, r_22, indexing="ij")
        np_l, np_z, np_r_33 = numpy.meshgrid(l, z, r_33, indexing="ij")
        
        np_r_12 = numpy.meshgrid(h, x, r_12, indexing="ij")[2]
        np_r_13 = numpy.meshgrid(k, y, r_13, indexing="ij")[2]
        np_r_23 = numpy.meshgrid(l, z, r_23, indexing="ij")[2]
        np_r_21 = numpy.meshgrid(h, x, r_21, indexing="ij")[2]
        np_r_31 = numpy.meshgrid(k, y, r_31, indexing="ij")[2]
        np_r_32 = numpy.meshgrid(l, z, r_32, indexing="ij")[2]

        np_b_1 = numpy.meshgrid(l, z, b_1, indexing="ij")[2]
        np_b_2 = numpy.meshgrid(l, z, b_2, indexing="ij")[2]
        np_b_3 = numpy.meshgrid(l, z, b_3, indexing="ij")[2]
        
        
        np_x_s = np_x*np_r_11 + np_y*np_r_12 + np_z*np_r_13 + np_b_1
        np_y_s = np_x*np_r_21 + np_y*np_r_22 + np_z*np_r_23 + np_b_2
        np_z_s = np_x*np_r_31 + np_y*np_r_32 + np_z*np_r_33 + np_b_3
        
        phase = numpy.exp(2*numpy.pi*1j*(np_h*np_x_s + np_k*np_y_s+ np_l*np_z_s))
        
        return phase
        
        
    def els4pos(self, space_group):
        """
        give the l_element of symmetry which transfer atom to the same atom
        """
        
        l_elsymm = space_group.el_symm
        l_orig = space_group.orig
        centr = space_group.centr
        pcentr = space_group.pcentr
    
        l_elsat = []
        l_els_uniq_at, l_coord_uniq_at = [], []
        [x, y, z] = xyz
        x, y, z = x%1, y%1, z%1
        for els in l_elsymm:
            for orig in l_orig:
                xat = (els[0] + els[1]*x + els[ 2]*y + els[ 3]*z+orig[0])%1
                yat = (els[4] + els[5]*x + els[ 6]*y + els[ 7]*z+orig[1])%1
                zat = (els[8] + els[9]*x + els[10]*y + els[11]*z+orig[2])%1
                elsn = [els[0]+orig[0],els[1],els[2],els[3],els[4]+orig[1],els[5],els[6],els[7],
                els[8]+orig[2],els[9],els[10],els[11]]
                if ((abs(xat-x)<10**-5)&(abs(yat-y)<10**-5)&(abs(zat-z)<10**-5)): l_elsat.append(elsn)
                xyzatu = (round(xat,4),round(yat,4),round(zat,4))
                if (not(xyzatu in l_coord_uniq_at)):
                    l_coord_uniq_at.append(xyzatu)
                    l_els_uniq_at.append(elsn)
                if (centr):
                    elsn=[2*pcentr[0]-els[0]-orig[0],-1*els[1],-1*els[2],-1*els[3],
                        2*pcentr[1]-els[4]-orig[1],-1*els[5],-1*els[6],-1*els[7],
                        2*pcentr[2]-els[8]-orig[2],-1*els[9],-1*els[10],-1*els[11]]
                    xat,yat,zat=(2*pcentr[0]-xat)%1,(2*pcentr[1]-yat)%1,(2*pcentr[2]-zat)%1
                    if ((abs(xat-x)<10**-5)&(abs(yat-y)<10**-5)&(abs(zat-z)<10**-5)): l_elsat.append(elsn)
                    xyzatu=(round(xat,4),round(yat,4),round(zat,4))
                    if (not(xyzatu in l_coord_uniq_at)):
                        l_coord_uniq_at.append(xyzatu)
                        l_els_uniq_at.append(elsn)
        return l_elsat,l_els_uniq_at




"""
internal class to calculate magnetic form factor
"""
__author__ = 'ikibalin'
__version__ = "2019_09_02"
import os
import numpy


def calc_mRmCmRT(r11, r12, r13, r21, r22, r23, r31, r32, r33,
                 c11, c12, c13, c21, c22, c23, c31, c32, c33):
    """
    calculate matrix multiplication R*C*RT, when matrices are expressed through 
    its component and can be expressed as nD-array
    """
    rc_11, rc_12 = r11*c11+r12*c21+r13*c31, r11*c12+r12*c22+r13*c32
    rc_13 = r11*c13+r12*c23+r13*c33
    rc_21, rc_22 = r21*c11+r22*c21+r23*c31, r21*c12+r22*c22+r23*c32
    rc_23 = r21*c13+r22*c23+r23*c33
    rc_31, rc_32 = r31*c11+r32*c21+r33*c31, r31*c12+r32*c22+r33*c32
    rc_33 = r31*c13+r32*c23+r33*c33

    #dimension (atoms, symmetry)
    rcrt_11 = (rc_11*r11+rc_12*r12+rc_13*r13)
    rcrt_12 = (rc_11*r21+rc_12*r22+rc_13*r23)
    rcrt_13 = (rc_11*r31+rc_12*r32+rc_13*r33)

    rcrt_21 = (rc_21*r11+rc_22*r12+rc_23*r13)
    rcrt_22 = (rc_21*r21+rc_22*r22+rc_23*r23)
    rcrt_23 = (rc_21*r31+rc_22*r32+rc_23*r33)

    rcrt_31 = (rc_31*r11+rc_32*r12+rc_33*r13)
    rcrt_32 = (rc_31*r21+rc_32*r22+rc_33*r23)
    rcrt_33 = (rc_31*r31+rc_32*r32+rc_33*r33)
    return rcrt_11, rcrt_12, rcrt_13, rcrt_21, rcrt_22, rcrt_23, rcrt_31, rcrt_32, rcrt_33



class Magnetism(object):
    """
    Magnetism
    """
    def __init__(self, factor_lande=numpy.array([], dtype=float),
                       kappa=numpy.array([], dtype=float),
                       chi_11=numpy.array([], dtype=float),
                       chi_12=numpy.array([], dtype=float),
                       chi_13=numpy.array([], dtype=float),
                       chi_22=numpy.array([], dtype=float),
                       chi_23=numpy.array([], dtype=float),
                       chi_33=numpy.array([], dtype=float),
                       j0_A=numpy.array([], dtype=float),
                       j0_a=numpy.array([], dtype=float),
                       j0_B=numpy.array([], dtype=float),
                       j0_b=numpy.array([], dtype=float),
                       j0_C=numpy.array([], dtype=float),
                       j0_c=numpy.array([], dtype=float),
                       j0_D=numpy.array([], dtype=float),
                       j2_A=numpy.array([], dtype=float),
                       j2_a=numpy.array([], dtype=float),
                       j2_B=numpy.array([], dtype=float),
                       j2_b=numpy.array([], dtype=float),
                       j2_C=numpy.array([], dtype=float),
                       j2_c=numpy.array([], dtype=float),
                       j2_D=numpy.array([], dtype=float)):
        super(Magnetism, self).__init__()
        self.__atom_site_aniso_magnetism_chi_11 = None
        self.__atom_site_aniso_magnetism_chi_12 = None
        self.__atom_site_aniso_magnetism_chi_13 = None
        self.__atom_site_aniso_magnetism_chi_22 = None
        self.__atom_site_aniso_magnetism_chi_23 = None
        self.__atom_site_aniso_magnetism_chi_33 = None
        self.__atom_site_magnetism_kappa = None
        self.__atom_site_magnetism_factor_lande = None
        self.__j0_A = None
        self.__j0_a = None
        self.__j0_B = None
        self.__j0_b = None
        self.__j0_C = None
        self.__j0_c = None
        self.__j0_D = None
        self.__j2_A = None
        self.__j2_a = None
        self.__j2_B = None
        self.__j2_b = None
        self.__j2_C = None
        self.__j2_c = None
        self.__j2_D = None

        self.__matrix_chi_loc = None

        self.factor_lande = factor_lande
        self.kappa = kappa
        self.chi_11 = chi_11
        self.chi_12 = chi_12
        self.chi_13 = chi_13
        self.chi_22 = chi_22
        self.chi_23 = chi_23
        self.chi_33 = chi_33
        self.j0_A = j0_A
        self.j0_a = j0_a
        self.j0_B = j0_B
        self.j0_b = j0_b
        self.j0_C = j0_C
        self.j0_c = j0_c
        self.j0_D = j0_D
        self.j2_A = j2_A
        self.j2_a = j2_a
        self.j2_B = j2_B
        self.j2_b = j2_b
        self.j2_C = j2_C
        self.j2_c = j2_c
        self.j2_D = j2_D

    def _trans_to_float_array(self, x):
        if isinstance(x, numpy.ndarray):
            x_out = x.astype(float)
        else:
            x_out = numpy.array([x], dtype=float)
        return x_out

    @property
    def factor_lande(self):
        return self.__atom_site_magnetism_factor_lande
    @factor_lande.setter
    def factor_lande(self, x):
        self.__atom_site_magnetism_factor_lande = self._trans_to_float_array(x)
    @property
    def kappa(self):
        return self.__atom_site_magnetism_kappa
    @kappa.setter
    def kappa(self, x):
        self.__atom_site_magnetism_kappa = self._trans_to_float_array(x)        
    @property
    def chi_11(self):
        return self.__atom_site_aniso_magnetism_chi_11
    @chi_11.setter
    def chi_11(self, x):
        self.__atom_site_aniso_magnetism_chi_11 = self._trans_to_float_array(x)    
    @property
    def chi_22(self):
        return self.__atom_site_aniso_magnetism_chi_22
    @chi_22.setter
    def chi_22(self, x):
        self.__atom_site_aniso_magnetism_chi_22 = self._trans_to_float_array(x)    
    @property
    def chi_33(self):
        return self.__atom_site_aniso_magnetism_chi_33
    @chi_33.setter
    def chi_33(self, x):
        self.__atom_site_aniso_magnetism_chi_33 = self._trans_to_float_array(x)    
    @property
    def chi_12(self):
        return self.__atom_site_aniso_magnetism_chi_12
    @chi_12.setter
    def chi_12(self, x):
        self.__atom_site_aniso_magnetism_chi_12 = self._trans_to_float_array(x)    
    @property
    def chi_13(self):
        return self.__atom_site_aniso_magnetism_chi_13
    @chi_13.setter
    def chi_13(self, x):
        self.__atom_site_aniso_magnetism_chi_13 = self._trans_to_float_array(x)    
    @property
    def chi_23(self):
        return self.__atom_site_aniso_magnetism_chi_23
    @chi_23.setter
    def chi_23(self, x):
        self.__atom_site_aniso_magnetism_chi_23 = self._trans_to_float_array(x)    

    @property
    def j0_A(self):
        return self.__j0_A
    @j0_A.setter
    def j0_A(self, x):
        self.__j0_A = self._trans_to_float_array(x)    
    @property
    def j0_a(self):
        return self.__j0_a
    @j0_a.setter
    def j0_a(self, x):
        self.__j0_a = self._trans_to_float_array(x)    
    @property
    def j0_B(self):
        return self.__j0_B
    @j0_B.setter
    def j0_B(self, x):
        self.__j0_B = self._trans_to_float_array(x)    
    @property
    def j0_b(self):
        return self.__j0_b
    @j0_b.setter
    def j0_b(self, x):
        self.__j0_b = self._trans_to_float_array(x)    
    @property
    def j0_C(self):
        return self.__j0_C
    @j0_C.setter
    def j0_C(self, x):
        self.__j0_C = self._trans_to_float_array(x)    
    @property
    def j0_c(self):
        return self.__j0_c
    @j0_c.setter
    def j0_c(self, x):
        self.__j0_c = self._trans_to_float_array(x)    
    @property
    def j0_D(self):
        return self.__j0_D
    @j0_D.setter
    def j0_D(self, x):
        self.__j0_D = self._trans_to_float_array(x)    


    @property
    def j2_A(self):
        return self.__j2_A
    @j2_A.setter
    def j2_A(self, x):
        self.__j2_A = self._trans_to_float_array(x)    
    @property
    def j2_a(self):
        return self.__j2_a
    @j2_a.setter
    def j2_a(self, x):
        self.__j2_a = self._trans_to_float_array(x)    
    @property
    def j2_B(self):
        return self.__j2_B
    @j2_B.setter
    def j2_B(self, x):
        self.__j2_B = self._trans_to_float_array(x)    
    @property
    def j2_b(self):
        return self.__j2_b
    @j2_b.setter
    def j2_b(self, x):
        self.__j2_b = self._trans_to_float_array(x)    
    @property
    def j2_C(self):
        return self.__j2_C
    @j2_C.setter
    def j2_C(self, x):
        self.__j2_C = self._trans_to_float_array(x)    
    @property
    def j2_c(self):
        return self.__j2_c
    @j2_c.setter
    def j2_c(self, x):
        self.__j2_c = self._trans_to_float_array(x)    
    @property
    def j2_D(self):
        return self.__j2_D
    @j2_D.setter
    def j2_D(self, x):
        self.__j2_D = self._trans_to_float_array(x)    

    @property
    def matrix_chi_loc(self):
        return self.__matrix_chi_loc

    def __repr__(self):
        ls_out = ["Magnetism susceptibility:\n  chi_11   chi_22   chi_33  chi_12   chi_13   chi_23"]
        ls_out.extend(["{:8.5f} {:8.5f} {:8.5f}{:8.5f} {:8.5f} {:8.5f}".format(
            hh_1, hh_2, hh_3, hh_4, hh_5, hh_6) for hh_1, hh_2, hh_3, hh_4, hh_5, hh_6 
            in zip(self.chi_11, self.chi_22, self.chi_33, self.chi_12, self.chi_13, self.chi_23)])
        return "\n".join(ls_out)

    
    def _calc_chi_loc(ia, ib, ic, matrix_ib):
        """
        IT SHOULD BE DELETED


        representation of chi in crystallographic coordinate system defined as x||a*, z||c, y= [z x] (right handed)
        expressions are taken from international tables
        matrix_ib is inversed matrix B
        ia, ib, ic is inversed unit cell parameters (it can be estimated from matrix matrix_ib)

        X = B x, x = iB X
        xT*CHI*x = XT iBT CHI iB X
    
        output chiLOC = iBT CHI iB
        """
        matrix_chi = numpy.array(
                [[self.chi_11, self.chi_12, self.chi_13],
                 [self.chi_12, self.chi_22, self.chi_23],
                 [self.chi_13, self.chi_23, self.chi_33]], 
                 dtype = float)
        #mchi=[[chi[0],chi[3],chi[4]],[chi[3],chi[1],chi[5]],[chi[4],chi[5],chi[2]]]
        #[a,b,c,alpha,beta,gamma]=ucp
        y1 = matrix_ib[0,0]
        y2 = matrix_ib[1,1]
        y3 = matrix_ib[2,2]
        y4 = matrix_ib[0,1]
        y5 = matrix_ib[0,2]
        y6 = matrix_ib[1,2]
        #B=[[x1,x4,x5],
        #   [0.,x2,x6],
        #   [0.,0.,x3]]
        #it shuld be checked
        #iB=numpy.linalg.inv(B)
        y1 = 1./x1
        y2 = 1./x2
        y3 = 1./x3
        y4 = -1*x4*1./(x1*x2)
        y6 = -1*x6*1./(x2*x3)
        y5 = (x4*x6-x2*x5)*1./(x1*x2*x3)
        matrix_ib_norm = matrix_ib
        #!!!! to check it !!! before it was [:, 0]
        matrix_ib_norm[0, :] *= ia
        matrix_ib_norm[1, :] *= ib
        matrix_ib_norm[2, :] *= ic
        print("There is mistake!!!!!")
        matrix_ibt_norm = matrix_ib_norm.transpose()
        #it is not compatible with case, vhen chi_ij is 1D array 
        ibt_chi = numpy.matmul(matrix_ibt_norm, matrix_chi)
        matrix_chi_loc = numpy.matmul(ibt_chi, matrix_ib_norm)
        self.__matrix_chi_loc = matrix_chi_loc
    
    def calc_form_factor_tensor(self, space_group, cell, h, k, l):
        """
        give components of form factor tensor:
            fft_11, fft_12, fft_13
            fft_21, fft_22, fft_23
            fft_31, fft_32, fft_33
            
        in 3 dimension (hkl, atoms, symmetry elements)            
        """
        sthovl = cell.calc_sthovl(h, k, l)
        #dimension (hkl, atoms)
        form_factor_2d = self._calc_form_factor(sthovl)
        
        r_11, r_12 = space_group.r_11, space_group.r_12
        r_13, r_21 = space_group.r_13, space_group.r_21
        r_22, r_23 = space_group.r_22, space_group.r_23
        r_31, r_32 = space_group.r_31, space_group.r_32
        r_33 = space_group.r_33

        chi_11, chi_22 = self.chi_11, self.chi_22 
        chi_33, chi_12 = self.chi_33, self.chi_12
        chi_13, chi_23 = self.chi_13, self.chi_23
        chi_21, chi_31, chi_32 = chi_12, chi_13, chi_23 
        
        c11, r11 = numpy.meshgrid(chi_11, r_11, indexing="ij")
        c22, r22 = numpy.meshgrid(chi_22, r_22, indexing="ij")
        c33, r33 = numpy.meshgrid(chi_33, r_33, indexing="ij")
        c12, r12 = numpy.meshgrid(chi_12, r_12, indexing="ij")
        c13, r13 = numpy.meshgrid(chi_13, r_13, indexing="ij")
        c23, r23 = numpy.meshgrid(chi_23, r_23, indexing="ij")
        c21, r21 = numpy.meshgrid(chi_21, r_21, indexing="ij")
        c31, r31 = numpy.meshgrid(chi_31, r_31, indexing="ij")
        c32, r32 = numpy.meshgrid(chi_32, r_32, indexing="ij")
        
        
        rcrt_11, rcrt_12, rcrt_13, rcrt_21, rcrt_22, rcrt_23, rcrt_31, rcrt_32, rcrt_33 = calc_mRmCmRT(
                r11, r12, r13, r21, r22, r23, r31, r32, r33,
                c11, c12, c13, c21, c22, c23, c31, c32, c33)        
        
        #dimension (hkl, atoms, symmetry)
        fft_11 = form_factor_2d[:,:,numpy.newaxis]*rcrt_11[numpy.newaxis, :,:]
        fft_12 = form_factor_2d[:,:,numpy.newaxis]*rcrt_12[numpy.newaxis, :,:]
        fft_13 = form_factor_2d[:,:,numpy.newaxis]*rcrt_13[numpy.newaxis, :,:]
        fft_21 = form_factor_2d[:,:,numpy.newaxis]*rcrt_21[numpy.newaxis, :,:]
        fft_22 = form_factor_2d[:,:,numpy.newaxis]*rcrt_22[numpy.newaxis, :,:]
        fft_23 = form_factor_2d[:,:,numpy.newaxis]*rcrt_23[numpy.newaxis, :,:]
        fft_31 = form_factor_2d[:,:,numpy.newaxis]*rcrt_31[numpy.newaxis, :,:]
        fft_32 = form_factor_2d[:,:,numpy.newaxis]*rcrt_32[numpy.newaxis, :,:]
        fft_33 = form_factor_2d[:,:,numpy.newaxis]*rcrt_33[numpy.newaxis, :,:]
        
        #ortogonalization should be done
        
        return fft_11, fft_12, fft_13, fft_21, fft_22, fft_23, fft_31, fft_32, fft_33
    
    def calc_chi_rot(self, matrix_chi, elsymm):
        """
        calculate R*chi*RT
        rotation of chi by element of symmetry
        """
        [b1,r11,r12,r13,b2,r21,r22,r23,b3,r31,r32,r33]=elsymm
        matrix_r = numpy.array([[r11, r12, r13], [r21, r22, r23], 
                               [r31, r32, r33]], dtype=float)
        matrix_rt = matrix_r.transpose()
        r_chi = numpy.matmul(matrix_r, matrix_chi)
        
        matrix_chi_rot = numpy.matmul(r_chi, matrix_rt)
        return matrix_chi_rot 
    
    def _calc_form_factor(self, sthovl):
        """Calculate magnetic form factor in frame of Spherical model (Int.Tabl.C.p.592)\n
        LFactor is Lande factor\n
        coeff0 is a list [A,a,B,b,C,c,D] at n=0\n
        coeff2 is a list [A,a,B,b,C,c,D] at n=2\n
        lsthovl is list sin(theta)/lambda in Angstrems**-1

        Calculation of magnetic form factor <j0>,<j2>,<j4>,<j6>\n
        coeff is a list [A,a,B,b,C,c,D] at n=0,2,4,6
        For help see International Table Vol.C p.460
        """
        #not sure about kappa, it is here just for test, by default it is 1.0
        kappa = self.kappa
        factor_lande = self.factor_lande
        j0_A = self.j0_A
        j0_a = self.j0_a
        j0_B = self.j0_B
        j0_b = self.j0_b
        j0_C = self.j0_C
        j0_c = self.j0_c
        j0_D = self.j0_D
        j2_A = self.j2_A
        j2_a = self.j2_a
        j2_B = self.j2_B
        j2_b = self.j2_b
        j2_C = self.j2_C
        j2_c = self.j2_c
        j2_D = self.j2_D     
        
        np_sthovl, np_kappa = numpy.meshgrid(sthovl, kappa, indexing ="ij")
        np_factor_lande = numpy.meshgrid(sthovl, factor_lande, indexing ="ij")[1]
        np_j0_A = numpy.meshgrid(sthovl, j0_A, indexing ="ij")[1]
        np_j0_a = numpy.meshgrid(sthovl, j0_a, indexing ="ij")[1]
        np_j0_B = numpy.meshgrid(sthovl, j0_B, indexing ="ij")[1]
        np_j0_b = numpy.meshgrid(sthovl, j0_b, indexing ="ij")[1]
        np_j0_C = numpy.meshgrid(sthovl, j0_C, indexing ="ij")[1]
        np_j0_c = numpy.meshgrid(sthovl, j0_c, indexing ="ij")[1]
        np_j0_D = numpy.meshgrid(sthovl, j0_D, indexing ="ij")[1]
        np_j2_A = numpy.meshgrid(sthovl, j2_A, indexing ="ij")[1]
        np_j2_a = numpy.meshgrid(sthovl, j2_a, indexing ="ij")[1]
        np_j2_B = numpy.meshgrid(sthovl, j2_B, indexing ="ij")[1]
        np_j2_b = numpy.meshgrid(sthovl, j2_b, indexing ="ij")[1]
        np_j2_C = numpy.meshgrid(sthovl, j2_C, indexing ="ij")[1]
        np_j2_c = numpy.meshgrid(sthovl, j2_c, indexing ="ij")[1]
        np_j2_D = numpy.meshgrid(sthovl, j2_D, indexing ="ij")[1]

        np_h = (np_sthovl*1./np_kappa)**2
        j0_av = (np_j0_A*numpy.exp(-np_j0_a*np_h)+
                 np_j0_B*numpy.exp(-np_j0_b*np_h)+
                 np_j0_C*numpy.exp(-np_j0_c*np_h)+np_j0_D)
        j2_av = (np_j2_A*numpy.exp(-np_j2_a*np_h)+
                 np_j2_B*numpy.exp(-np_j2_b*np_h)+
                 np_j2_C*numpy.exp(-np_j2_c*np_h)+np_j2_D)*np_h
        #according https://journals.aps.org/prb/pdf/10.1103/PhysRevB.79.140405
        #mismatch with international tables where (1.0-2.0/np_factor_lande)
        form_factor = j0_av+(2.0/np_factor_lande-1.0)*j2_av
        return form_factor




"""
define classe Pd2dProc for 1d powder diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_10"
import os
import numpy


from pycifstar import Global


class Pd2dProc(object):
    """
    This section contains the diffraction data set after processing
    and application of correction terms. 

    Example:

    _pd2d_proc_2theta_phi_intensity_up_net
    ;
    ;
    _pd2d_proc_2theta_phi_intensity_down_net
    ;
    ;
    _pd2d_proc_2theta_phi_intensity_up_total
    ;
    ;
    _pd2d_proc_2theta_phi_intensity_down_total
    ;
    ;
    _pd2d_proc_2theta_phi_intensity_bkg_calc
    ;
    ;
    _pd2d_proc_2theta_phi_intensity_up
    ;
    ;
    _pd2d_proc_2theta_phi_intensity_up_sigma
    ;
    ;
    _pd2d_proc_2theta_phi_intensity_down
    ;
    ;
    _pd2d_proc_2theta_phi_intensity_down_sigma
    ;
    ;
    """
    def __init__(self, ttheta=[], phi=[],
                up_net=[[]], down_net=[[]], up_total=[[]], down_total=[[]], bkg_calc=[[]], 
                up=[[]], down=[[]], up_sigma=[[]], down_sigma=[[]]):
        super(Pd2dProc, self).__init__()
        self.__pd2d_proc_2theta = None
        self.__pd2d_proc_phi = None
        self.__pd2d_proc_intensity_up_net = None
        self.__pd2d_proc_intensity_down_net = None
        self.__pd2d_proc_intensity_up_total = None
        self.__pd2d_proc_intensity_down_total = None
        self.__pd2d_proc_intensity_bkg_calc = None
        self.__pd2d_proc_intensity_up = None
        self.__pd2d_proc_intensity_up_sigma = None
        self.__pd2d_proc_intensity_down = None
        self.__pd2d_proc_intensity_down_sigma = None


        self.ttheta = ttheta
        self.phi = phi
        self.up_net = up_net
        self.down_net = down_net
        self.up_total = up_total
        self.down_total = down_total
        self.bkg_calc = bkg_calc
        self.up = up
        self.up_sigma = up_sigma
        self.down = down
        self.down_sigma = down_sigma

    @property
    def ttheta(self):
        return self.__pd2d_proc_2theta
    @ttheta.setter
    def ttheta(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd2d_proc_2theta = np_x_in


    @property
    def phi(self):
        return self.__pd2d_proc_phi
    @phi.setter
    def phi(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd2d_proc_phi = np_x_in


    @property
    def up_net(self):
        return self.__pd2d_proc_intensity_up_net
    @up_net.setter
    def up_net(self, ll_x):
        ll_x_in = []
        for l_x in ll_x:
            l_x_in = []
            for x in l_x:
                if (isinstance(x, float) | (x is None)):
                    x_in = x
                else:
                    x_in = float(x)
                l_x_in.append(x_in)
            ll_x_in.append(l_x_in)
        np_x_in = numpy.array(ll_x_in, dtype=float)
        self.__pd2d_proc_intensity_up_net = np_x_in


    @property
    def down_net(self):
        return self.__pd2d_proc_intensity_down_net
    @down_net.setter
    def down_net(self, ll_x):
        ll_x_in = []
        for l_x in ll_x:
            l_x_in = []
            for x in l_x:
                if (isinstance(x, float) | (x is None)):
                    x_in = x
                else:
                    x_in = float(x)
                l_x_in.append(x_in)
            ll_x_in.append(l_x_in)
        np_x_in = numpy.array(ll_x_in, dtype=float)
        self.__pd2d_proc_intensity_down_net = np_x_in


    @property
    def up_total(self):
        return self.__pd2d_proc_intensity_up_total
    @up_total.setter
    def up_total(self, ll_x):
        ll_x_in = []
        for l_x in ll_x:
            l_x_in = []
            for x in l_x:
                if (isinstance(x, float) | (x is None)):
                    x_in = x
                else:
                    x_in = float(x)
                l_x_in.append(x_in)
            ll_x_in.append(l_x_in)
        np_x_in = numpy.array(ll_x_in, dtype=float)
        self.__pd2d_proc_intensity_up_total = np_x_in


    @property
    def down_total(self):
        return self.__pd2d_proc_intensity_down_total
    @down_total.setter
    def down_total(self, ll_x):
        ll_x_in = []
        for l_x in ll_x:
            l_x_in = []
            for x in l_x:
                if (isinstance(x, float) | (x is None)):
                    x_in = x
                else:
                    x_in = float(x)
                l_x_in.append(x_in)
            ll_x_in.append(l_x_in)
        np_x_in = numpy.array(ll_x_in, dtype=float)
        self.__pd2d_proc_intensity_down_total = np_x_in


    @property
    def bkg_calc(self):
        return self.__pd2d_proc_intensity_bkg_calc
    @bkg_calc.setter
    def bkg_calc(self, ll_x):
        ll_x_in = []
        for l_x in ll_x:
            l_x_in = []
            for x in l_x:
                if (isinstance(x, float) | (x is None)):
                    x_in = x
                else:
                    x_in = float(x)
                l_x_in.append(x_in)
            ll_x_in.append(l_x_in)
        np_x_in = numpy.array(ll_x_in, dtype=float)
        self.__pd2d_proc_intensity_bkg_calc = np_x_in


    @property
    def up(self):
        return self.__pd2d_meas_intensity_up
    @up.setter
    def up(self, ll_x):
        ll_x_in = []
        for l_x in ll_x:
            l_x_in = []
            for x in l_x:
                if (isinstance(x, float) | (x is None)):
                    x_in = x
                else:
                    x_in = float(x)
                l_x_in.append(x_in)
            ll_x_in.append(l_x_in)
        np_x_in = numpy.array(ll_x_in, dtype=float)
        self.__pd2d_meas_intensity_up = np_x_in

    @property
    def up_sigma(self):
        return self.__pd2d_meas_intensity_up_sigma
    @up_sigma.setter
    def up_sigma(self, ll_x):
        ll_x_in = []
        for l_x in ll_x:
            l_x_in = []
            for x in l_x:
                if (isinstance(x, float) | (x is None)):
                    x_in = x
                else:
                    x_in = float(x)
                l_x_in.append(x_in)
            ll_x_in.append(l_x_in)
        np_x_in = numpy.array(ll_x_in, dtype=float)
        self.__pd2d_meas_intensity_up_sigma = np_x_in

    @property
    def down(self):
        return self.__pd2d_meas_intensity_down
    @down.setter
    def down(self, ll_x):
        ll_x_in = []
        for l_x in ll_x:
            l_x_in = []
            for x in l_x:
                if (isinstance(x, float) | (x is None)):
                    x_in = x
                else:
                    x_in = float(x)
                l_x_in.append(x_in)
            ll_x_in.append(l_x_in)
        np_x_in = numpy.array(ll_x_in, dtype=float)
        self.__pd2d_meas_intensity_down = np_x_in

    @property
    def down_sigma(self):
        return self.__pd2d_meas_intensity_down_sigma
    @down_sigma.setter
    def down_sigma(self, ll_x):
        ll_x_in = []
        for l_x in ll_x:
            l_x_in = []
            for x in l_x:
                if (isinstance(x, float) | (x is None)):
                    x_in = x
                else:
                    x_in = float(x)
                l_x_in.append(x_in)
            ll_x_in.append(l_x_in)
        np_x_in = numpy.array(ll_x_in, dtype=float)
        self.__pd2d_meas_intensity_down_sigma = np_x_in


            
    def __repr__(self):
        ls_out = ["Pd2dProc:"]
        ls_out.append("\n ttheta: min - {:}°, max - {:}°, points - {:}".format(self.ttheta.min(), self.ttheta.max(), self.ttheta.size))
        ls_out.append(" phi:    min - {:}°, max - {:}°, points - {:}".format(self.phi.min(), self.phi.max(), self.phi.size))
        ls_out.append(" phi: {:}".format(str(self.phi)))
        ls_out.append("\n up_total")
        ls_out.append(str(self.up_total.transpose()))
        ls_out.append("\n up")
        ls_out.append(str(self.up.transpose()))
        ls_out.append("\n down_total")
        ls_out.append(str(self.down_total.transpose()))
        ls_out.append("\n down")
        ls_out.append(str(self.down.transpose()))
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("_pd2d_proc_2theta_phi_intensity_up_net")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            for phi, l_intensity in zip(self.phi, self.up_net.transpose()):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) if _ is not numpy.nan else "        None" for _ in l_intensity]))
            ls_out.append(";")

            ls_out.append("_pd2d_proc_2theta_phi_intensity_down_net")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            for phi, l_intensity in zip(self.phi, self.down_net.transpose()):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) if _ is not numpy.nan else "        None" for _ in l_intensity]))
            ls_out.append(";")

            ls_out.append("_pd2d_proc_2theta_phi_intensity_up_total")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            for phi, l_intensity in zip(self.phi, self.up_total.transpose()):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) if _ is not numpy.nan else "        None" for _ in l_intensity]))
            ls_out.append(";")

            ls_out.append("_pd2d_proc_2theta_phi_intensity_down_total")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            for phi, l_intensity in zip(self.phi, self.down_total.transpose()):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) if _ is not numpy.nan else "        None" for _ in l_intensity]))
            ls_out.append(";")

            ls_out.append("_pd2d_proc_2theta_phi_intensity_bkg_calc")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            for phi, l_intensity in zip(self.phi, self.bkg_calc.transpose()):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) if _ is not numpy.nan else "        None" for _ in l_intensity]))
            ls_out.append(";")

            ls_out.append("_pd2d_proc_2theta_phi_intensity_up")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            for phi, l_intensity in zip(self.phi, self.up.transpose()):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) if _ is not numpy.nan else "        None" for _ in l_intensity]))
            ls_out.append(";")

            ls_out.append("_pd2d_proc_2theta_phi_intensity_up_sigma")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            for phi, l_intensity in zip(self.phi, self.up_sigma.transpose()):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) if _ is not numpy.nan else "        None" for _ in l_intensity]))
            ls_out.append(";")

            ls_out.append("_pd2d_proc_2theta_phi_intensity_down")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            for phi, l_intensity in zip(self.phi, self.down.transpose()):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) if _ is not numpy.nan else "        None" for _ in l_intensity]))
            ls_out.append(";")

            ls_out.append("_pd2d_proc_2theta_phi_intensity_down_sigma")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            for phi, l_intensity in zip(self.phi, self.down_sigma.transpose()):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) if _ is not numpy.nan else "        None" for _ in l_intensity]))
            ls_out.append(";")
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = cif_global.is_prefix("_pd2d_proc_2theta_phi_intensity_up_net")
        if flag:
            cif_value = cif_global["_pd2d_proc_2theta_phi_intensity_up_net"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.up_net = ll_intensity

        flag = cif_global.is_prefix("_pd2d_proc_2theta_phi_intensity_down_net")
        if flag:
            cif_value = cif_global["_pd2d_proc_2theta_phi_intensity_down_net"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.down_net = ll_intensity

        flag = cif_global.is_prefix("_pd2d_proc_2theta_phi_intensity_up_total")
        if flag:
            cif_value = cif_global["_pd2d_proc_2theta_phi_intensity_up_total"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.up_total = ll_intensity

        flag = cif_global.is_prefix("_pd2d_proc_2theta_phi_intensity_down_total")
        if flag:
            cif_value = cif_global["_pd2d_proc_2theta_phi_intensity_down_total"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.down_total = ll_intensity


        flag = cif_global.is_prefix("_pd2d_proc_2theta_phi_intensity_bkg_calc")
        if flag:
            cif_value = cif_global["_pd2d_proc_2theta_phi_intensity_bkg_calc"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.bkg_calc = ll_intensity

        flag = cif_global.is_prefix("_pd2d_proc_2theta_phi_intensity_up")
        if flag:
            cif_value = cif_global["_pd2d_proc_2theta_phi_intensity_up"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.up = ll_intensity

        flag = cif_global.is_prefix("_pd2d_proc_2theta_phi_intensity_up_sigma")
        if flag:
            cif_value = cif_global["_pd2d_proc_2theta_phi_intensity_up_sigma"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.up_sigma = ll_intensity

        flag = cif_global.is_prefix("_pd2d_proc_2theta_phi_intensity_down")
        if flag:
            cif_value = cif_global["_pd2d_proc_2theta_phi_intensity_down"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.down = ll_intensity

        flag = cif_global.is_prefix("_pd2d_proc_2theta_phi_intensity_down_sigma")
        if flag:
            cif_value = cif_global["_pd2d_proc_2theta_phi_intensity_down_sigma"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.down_sigma = ll_intensity
        return True

    @property
    def is_defined(self):
        cond = all([self.ttheta is not None, self.phi is not None, 
                    self.up_net is not None, self.down_net is not None, self.up_total is not None, self.down_total is not None,
                    self.bkg_calc is not None,
                    self.up is not None, self.up_sigma is not None, self.down is not None, self.down_sigma is not None])
        return cond

    @property
    def is_variable(self):
        return False
    
    def get_variables(self):
        return []

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)
